#include <iostream>
#include <iomanip>
#include <sys/stat.h>
#include <fstream>
#include <string>
#include <cstring>
#include <io.h>
#include <dirent.h>
#include <direct.h>
#include <zlib.h>
#include "blowfish.cpp"
#include "RFFTool.h"

using namespace std;

#define ceiling8(x) (((x) & 7) ? (((x) & (~7)) + 8) : (x))

unsigned char REALKEY [8] = {0x72, 0x03, 0x17, 0xF4, 0x12, 0x43, 0x92, 0x88};
unsigned char DEMOKEY[8] = {0x23, 0x43, 0x55, 0x02, 0x37, 0x99, 0xaf, 0xc7};
unsigned char *KEY = &REALKEY[0];


char *ArchiveFilename, *Directoryname;
int sw_unc = 0, sw_list = 0, sw_comp = 0, sw_add = 0, sw_dircomp = 0;

void makedirs (char *FileName)
{
	char *Path = FileName;
	char Dir;
	while (Path = strchr (Path, '/')) {
		Dir = *Path;
		*Path = 0;
		if (_access (FileName, 0))
			_mkdir (FileName);
		*Path++ = Dir;
	}
}

void error_msg (string msg, int errorcode=1, char *s=NULL)
{
	if (s) msg.replace (msg.find ("%arg"), 4, s);
	cout << msg << endl;
	exit (errorcode);
}

void deflate (void)
{
	char ID[32];
	unsigned long ArchiveType, IndexOffset, FileCount, ArchiveSize, pos;
	RFF_subfile subfile;

	ifstream in (ArchiveFilename, ios::binary);
	if (!in) error_msg (MSG_OPEN, 1, ArchiveFilename);

	in.seekg (0, ios::end);	// Find archive size
	ArchiveSize = in.tellg ();
	if (ArchiveSize < 36) error_msg (MSG_NOTVALID, 1, ArchiveFilename);
	in.seekg (0);

	in.read (ID, 32); //Check file ID
	if (strcmp (ID, "Refractor2 FinkelArchive 1.0"))
		error_msg (MSG_NOTVALID, 1, ArchiveFilename);
	in.read (reinterpret_cast<char*>(&IndexOffset), 4);
	in.read (reinterpret_cast<char*>(&ArchiveType), 4);
	if (!in) error_msg (in.eof () ? MSG_EOF : MSG_READ);
	if (ArchiveType > 2) error_msg (MSG_TYPE, 1, ArchiveFilename);

    cout << "ArchiveType\t\t: " << ArchiveType;
    if (ArchiveType == 0) cout << " (Straight forward)" << endl;
    if (ArchiveType == 1) cout << " (Encrypted, not compressed)" << endl;
    if (ArchiveType == 2) cout << " (Encrypted and compressed)" << endl;

	if (IndexOffset + 4 > ArchiveSize) error_msg (MSG_OFFSET);

	in.seekg (IndexOffset);
	in.read (reinterpret_cast<char*>(&FileCount), 4);
	if (!in) error_msg (in.eof () ? MSG_EOF : MSG_READ);

	cout << "FileCount\t\t: "<< FileCount << endl << endl;
	if (!FileCount) error_msg (MSG_NOFILES);

	cout << "FileName                                                         CSize   USize\n";

	for (unsigned int i = 0; i < FileCount; i++) {
		in.read (reinterpret_cast<char*>(&subfile.FileNameLength), 4);
		subfile.FileName = new char [subfile.FileNameLength + 1];
		subfile.FileName [subfile.FileNameLength] = '\0';
		in.read (subfile.FileName, subfile.FileNameLength);
		in.read (reinterpret_cast<char*>(&subfile.Size), 16);
		if (!in) error_msg (in.eof () ? MSG_EOF : MSG_READ);

		pos = in.tellg () + 8; // Set pos to next file header...

		if (sw_unc) {
			in.seekg (subfile.DataOffset);
			subfile.CData = new char [ceiling8 (subfile.Size)];
			if (ArchiveType == 2) subfile.UData = new char [subfile.UncompressedSize];

			in.read (subfile.CData, ceiling8 (subfile.Size));
			if (ArchiveType > 0) Blowfish_DecryptBlock (subfile.CData, ceiling8 (subfile.Size), KEY, 8);
			if (ArchiveType < 2) subfile.UData = subfile.CData;
			if (ArchiveType == 2) uncompress (subfile.UData, &subfile.UncompressedSize, subfile.CData, subfile.Size);
			makedirs (subfile.FileName);
			ofstream out(subfile.FileName, ios::binary);
			if (!out) error_msg (MSG_CREATE, 1, subfile.FileName);
			out.write (subfile.UData, subfile.UncompressedSize);
			if (!out) error_msg (MSG_WRITE, 1, subfile.FileName);
			out.close ();
			delete[] subfile.CData;
			if (ArchiveType == 2) delete[] subfile.UData;
		}

		in.seekg (pos);

		cout << setw (62) << setiosflags (ios::left) << subfile.FileName << resetiosflags (ios::left)
		<< setw (8) << subfile.Size << setw (8) << subfile.UncompressedSize << endl;

		delete[] subfile.FileName;
	}

	in.close();

	cout << endl;
	if (!sw_unc && sw_list) cout << "Files listed, not uncompressed!\n";
	if (sw_unc) cout << "Files successfully uncompressed!\n";
}

int countfiles (char *dirname)
{
	DIR *handle;
	struct stat finfo;
	dirent *direntry;
	int count = 0;
	char temp[256];

	if (!(handle = opendir (dirname)))
		error_msg (MSG_OPENDIR, 1, dirname);

	while (direntry = readdir (handle)) {
		if (direntry->d_name[0] == '.')
			continue;
		strcpy (temp, dirname);
		strcat (temp, "\\");
		strcat (temp, direntry->d_name);
		stat (temp, &finfo);
		if (finfo.st_mode & S_IFDIR) count += countfiles (temp);
		else count++;
	}
	closedir (handle);
	return count;
}

int makefilelist (char *dirname, RFF_subfile *filelist, unsigned long filecount, unsigned long count=0)
{
	DIR *handle;
	struct stat finfo;
	dirent *direntry;
	char temp[256];

	if (!(handle = opendir (dirname)))
		error_msg (MSG_OPENDIR, 1, dirname);

	while (direntry = readdir (handle)) {
		if (direntry->d_name[0] == '.')
			continue;
		strcpy (temp, dirname);
		strcat (temp, "\\");
		strcat (temp, direntry->d_name);
		stat (temp, &finfo);
		if (finfo.st_mode & S_IFDIR)
			count = makefilelist (temp, filelist, filecount, count);
		else {
			filelist [count].FileNameLength = strlen (temp);
			filelist [count].FileName = new char [filelist [count].FileNameLength + 1];
			strcpy (filelist [count].FileName, temp);
			filelist [count].UncompressedSize = finfo.st_size;
			filelist [count].Type = 1;
			count++;
		}
	}
	closedir (handle);
	return count;
}

void inflate (void)
{
	unsigned long ArchiveType = 2, IndexOffset, FileCount, pos;
	int Z_status;
	RFF_subfile *subfile;

	FileCount = countfiles (Directoryname);
	if (FileCount == 0) error_msg (MSG_NOFILESFOUND);
	cout << "ArchiveType\t\t: " << ArchiveType;
	if (ArchiveType == 0) cout << " (Straight forward)" << endl;
	if (ArchiveType == 1) cout << " (Encrypted, not compressed)" << endl;
    if (ArchiveType == 2) cout << " (Encrypted and compressed)" << endl;
	cout << "FileCount\t\t: " << FileCount << endl << endl;
	subfile = new RFF_subfile [FileCount];
	makefilelist (Directoryname, subfile, FileCount);

	fstream out (ArchiveFilename, ios::binary | ios::out);
	if (!out) error_msg (MSG_OPEN, 1, ArchiveFilename);

	out.write ("Refractor2 FinkelArchive 1.0\0\0\0\0\0\0\0\0", 36);
	out.write (reinterpret_cast<char*>(&ArchiveType), 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	pos = 40;

	cout << "FileName                                                         USize   CSize\n";
	for (unsigned long i = 0; i < FileCount; i++) {
		subfile [i].DataOffset = pos;
		subfile [i].UData = new char [subfile [i].UncompressedSize];
		ifstream in (subfile [i].FileName, ios::binary);
		if (!in) error_msg (MSG_OPEN, 1, subfile [i].FileName);
		in.read (subfile [i].UData, subfile [i].UncompressedSize);
		if (!in) error_msg (MSG_READ, 1, subfile [i].FileName);
		in.close ();

		for (unsigned long j = 0; j < subfile [i].FileNameLength; j++)
					if (subfile [i].FileName [j] == '\\') subfile [i].FileName [j] = '/';

		cout << setw (62) << setiosflags (ios::left) << subfile [i].FileName << resetiosflags (ios::left) << setw (8) << subfile [i].UncompressedSize;

		subfile [i].Size = subfile [i].UncompressedSize * 2 + 1024;
		subfile [i].CData = new char [subfile [i].Size];
		if ((Z_status = compress (subfile [i].CData, &subfile [i].Size, subfile [i].UData, subfile [i].UncompressedSize)) != Z_OK) {
			if (Z_status == Z_MEM_ERROR) error_msg (MSG_MEMORY);
			else error_msg (MSG_COMPRESS);
		}

		if (subfile [i].Size % 8)
			for (unsigned long j = 0; j < 8 - (subfile [i].Size % 8); j++)
				subfile [i].CData [subfile [i].Size + j] = 0; // Pad with 0x00

		Blowfish_EncryptBlock (subfile [i].CData, ceiling8 (subfile[i].Size), KEY, 8);
		out.write (subfile [i].CData, ceiling8 (subfile [i].Size));
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
		pos += ceiling8 (subfile [i].Size);
		cout << setw (8) << subfile [i].Size << endl;

		delete [] subfile [i].UData;
		delete [] subfile [i].CData;
	}

	IndexOffset = pos;
	out.write (reinterpret_cast<char*>(&FileCount), 4);
	for (unsigned long i = 0; i < FileCount; i++) {
		out.write (reinterpret_cast<char*>(&subfile [i].FileNameLength), 4);
		out.write (subfile [i].FileName, subfile [i].FileNameLength);
		out.write (reinterpret_cast<char*>(&subfile [i].Size), 16);
		out.write ("\0\0\0\0\0\0\0\0", 8);
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	}
	out.write ("\0\0\0\0", 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.seekg (32);
	out.write (reinterpret_cast<char*>(&IndexOffset), 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.close ();

	for (unsigned long i = 0; i < FileCount; i++)
		delete [] subfile [i].FileName;
	delete [] subfile;

	cout << "\nFiles successfully compressed!\n";
}

void add (void)
{
	unsigned long ArchiveType = 2, IndexOffset, FileCount, OldCount, TotalCount, pos;
	int Z_status;
	char ID[32];
	RFF_subfile *subfile;
	ifstream in;

	FileCount = countfiles (Directoryname);
	if (FileCount == 0) error_msg (MSG_NOFILESFOUND);
	in.open (ArchiveFilename, ios::binary | ios::in);
	if (!in) error_msg (MSG_OPEN, 1, ArchiveFilename);
	in.read (ID, 32); //Check file ID
	if (strcmp (ID, "Refractor2 FinkelArchive 1.0"))
		error_msg (MSG_NOTVALID, 1, ArchiveFilename);
	in.read (reinterpret_cast<char*>(&IndexOffset), 4);
	in.seekg (IndexOffset);
	in.read (reinterpret_cast<char*>(&OldCount), 4);
	cout << "got here";
	if (!in) error_msg (in.eof () ? MSG_EOF : MSG_READ);

	TotalCount = FileCount + OldCount;
	subfile = new RFF_subfile [TotalCount];
	cout << "got here";
	for (unsigned long i = 0; i < OldCount; i++) {
		in.read (reinterpret_cast<char*>(&subfile [i].FileNameLength), 4);
		subfile [i].FileName = new char [subfile [i].FileNameLength + 1];
		subfile [i].FileName [subfile [i].FileNameLength] = '\0';
		in.read (subfile [i].FileName, subfile [i].FileNameLength);
		in.read (reinterpret_cast<char*>(&subfile [i].Size), 16);
		if (!in) error_msg (in.eof () ? MSG_EOF : MSG_READ);
		in.seekg (8, ios::cur);
	}
	in.close ();

	cout << "ArchiveType\t\t: " << ArchiveType;
	if (ArchiveType == 0) cout << " (Straight forward)" << endl;
	if (ArchiveType == 1) cout << " (Encrypted, not compressed)" << endl;
    if (ArchiveType == 2) cout << " (Encrypted and compressed)" << endl;
	cout << "Files to add\t: " << FileCount << endl << endl;

	makefilelist (Directoryname, subfile, FileCount, OldCount);
	fstream out (ArchiveFilename, ios::binary | ios::out | ios::in);
	out.seekg (IndexOffset);
	pos = IndexOffset;

	cout << "FileName                                                         USize   CSize\n";
	for (unsigned long i = OldCount; i < FileCount + OldCount; i++) {
		subfile [i].DataOffset = pos;
		subfile [i].UData = new char [subfile [i].UncompressedSize];
		in.open (subfile [i].FileName, ios::binary);
		if (!in) error_msg (MSG_OPEN, 1, subfile [i].FileName);
		in.read (subfile [i].UData, subfile [i].UncompressedSize);
		if (!in) error_msg (MSG_READ, 1, subfile [i].FileName);
		in.close ();

		for (unsigned long j = 0; j < subfile [i].FileNameLength; j++)
					if (subfile [i].FileName [j] == '\\') subfile [i].FileName [j] = '/';

		cout << setw (62) << setiosflags (ios::left) << subfile [i].FileName << resetiosflags (ios::left) << setw (8) << subfile [i].UncompressedSize;

		subfile [i].Size = subfile [i].UncompressedSize * 2 + 1024;
		subfile [i].CData = new char [subfile [i].Size];
		if ((Z_status = compress (subfile [i].CData, &subfile [i].Size, subfile [i].UData, subfile [i].UncompressedSize)) != Z_OK) {
			if (Z_status == Z_MEM_ERROR) error_msg (MSG_MEMORY);
			else error_msg (MSG_COMPRESS);
		}

		if (subfile [i].Size % 8)
			for (unsigned long j = 0; j < 8 - (subfile [i].Size % 8); j++)
				subfile [i].CData [subfile [i].Size + j] = 0; // Pad with 0x00

		Blowfish_EncryptBlock (subfile [i].CData, ceiling8 (subfile[i].Size), KEY, 8);
		out.write (subfile [i].CData, ceiling8 (subfile [i].Size));
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
		pos += ceiling8 (subfile [i].Size);
		cout << setw (8) << subfile [i].Size << endl;

		delete [] subfile [i].UData;
		delete [] subfile [i].CData;
	}

	IndexOffset = pos;
	out.write (reinterpret_cast<char*>(&TotalCount), 4);
	for (unsigned long i = 0; i < TotalCount; i++) {
		out.write (reinterpret_cast<char*>(&subfile [i].FileNameLength), 4);
		out.write (subfile [i].FileName, subfile [i].FileNameLength);
		out.write (reinterpret_cast<char*>(&subfile [i].Size), 16);
		out.write ("\0\0\0\0\0\0\0\0", 8);
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	}
	out.write ("\0\0\0\0", 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.seekg (32);
	out.write (reinterpret_cast<char*>(&IndexOffset), 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.close ();

	for (unsigned long i = 0; i < FileCount; i++)
		delete [] subfile [i].FileName;
	delete [] subfile;

	cout << "\nFiles successfully added!\n";
}

void inflatedir (void)
{
	unsigned long ArchiveType = 2, IndexOffset, FileCount, DirnameLength, pos, temp;
	int Z_status;
	RFF_subfile *subfile;
	DirnameLength = strlen (Directoryname) + 1;

	FileCount = countfiles (Directoryname);
	if (FileCount == 0) error_msg (MSG_NOFILESFOUND);
	cout << "ArchiveType\t\t: " << ArchiveType;
	if (ArchiveType == 0) cout << " (Straight forward)" << endl;
	if (ArchiveType == 1) cout << " (Encrypted, not compressed)" << endl;
    if (ArchiveType == 2) cout << " (Encrypted and compressed)" << endl;
	cout << "FileCount\t\t: " << FileCount << endl << endl;
	subfile = new RFF_subfile [FileCount];
	makefilelist (Directoryname, subfile, FileCount);

	fstream out (ArchiveFilename, ios::binary | ios::out);
	if (!out) error_msg (MSG_OPEN, 1, ArchiveFilename);

	out.write ("Refractor2 FinkelArchive 1.0\0\0\0\0\0\0\0\0", 36);
	out.write (reinterpret_cast<char*>(&ArchiveType), 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	pos = 40;

	cout << "FileName                                                         USize   CSize\n";
	for (unsigned long i = 0; i < FileCount; i++) {
		subfile [i].DataOffset = pos;
		subfile [i].UData = new char [subfile [i].UncompressedSize];
		ifstream in (subfile [i].FileName, ios::binary);
		if (!in) error_msg (MSG_OPEN, 1, subfile [i].FileName);
		in.read (subfile [i].UData, subfile [i].UncompressedSize);
		if (!in) error_msg (MSG_READ, 1, subfile [i].FileName);
		in.close ();

		for (unsigned long j = 0; j < subfile [i].FileNameLength; j++)
					if (subfile [i].FileName [j] == '\\') subfile [i].FileName [j] = '/';

		cout << setw (62) << setiosflags (ios::left) << &subfile [i].FileName [DirnameLength] << resetiosflags (ios::left) << setw (8) << subfile [i].UncompressedSize;

		subfile [i].Size = subfile [i].UncompressedSize * 2 + 1024;
		subfile [i].CData = new char [subfile [i].Size];
		if ((Z_status = compress (subfile [i].CData, &subfile [i].Size, subfile [i].UData, subfile [i].UncompressedSize)) != Z_OK) {
			if (Z_status == Z_MEM_ERROR) error_msg (MSG_MEMORY);
			else error_msg (MSG_COMPRESS);
		}

		if (subfile [i].Size % 8)
			for (unsigned long j = 0; j < 8 - (subfile [i].Size % 8); j++)
				subfile [i].CData [subfile [i].Size + j] = 0; // Pad with 0x00

		Blowfish_EncryptBlock (subfile [i].CData, ceiling8 (subfile[i].Size), KEY, 8);
		out.write (subfile [i].CData, ceiling8 (subfile [i].Size));
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
		pos += ceiling8 (subfile [i].Size);
		cout << setw (8) << subfile [i].Size << endl;

		delete [] subfile [i].UData;
		delete [] subfile [i].CData;
	}

	IndexOffset = pos;
	out.write (reinterpret_cast<char*>(&FileCount), 4);
	for (unsigned long i = 0; i < FileCount; i++) {
		unsigned long temp = subfile [i].FileNameLength - DirnameLength;
		out.write (reinterpret_cast<char*>(&temp), 4);
		out.write (&subfile [i].FileName [DirnameLength], temp);
		out.write (reinterpret_cast<char*>(&subfile [i].Size), 16);
		out.write ("\0\0\0\0\0\0\0\0", 8);
		if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	}
	out.write ("\0\0\0\0", 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.seekg (32);
	out.write (reinterpret_cast<char*>(&IndexOffset), 4);
	if (!out) error_msg (MSG_WRITE, 1, ArchiveFilename);
	out.close ();

	for (unsigned long i = 0; i < FileCount; i++)
		delete [] subfile [i].FileName;
	delete [] subfile;

	cout << "\nFiles successfully compressed!\n";
}

int main (int argc, char *argv[])
{
	cout << "RFFTool ver. 1.0 by Inside Knowledge\n\n";

	char *SwitchStr;

	if (argc < 2 || argc > 4) error_msg (MSG_USAGE);
	if (argc > 2) {
		SwitchStr = argv [1];
		if (*SwitchStr == '-') {
			while (*(++SwitchStr) != '\0') {
				if (*SwitchStr == 'a') sw_add = 1;
				if (*SwitchStr == 'c') sw_comp = 1;
				if (*SwitchStr == 'l') sw_list = 1;
				if (*SwitchStr == 'd') {
					sw_unc = 1;
					sw_list = 1;
				}
				if (*SwitchStr == 'q') KEY = &DEMOKEY[0];
			}
			if (!sw_list && !sw_unc && !sw_comp && !sw_add) error_msg (MSG_USAGE);
			ArchiveFilename = argv[2];
			if (sw_comp || sw_add) {
				if (argc == 4) Directoryname = argv [3];
				else error_msg (MSG_USAGE);
			}
			else if (argc == 4) error_msg (MSG_USAGE);
		}
		else error_msg (MSG_USAGE);
	}

	if (argc == 2) {
		struct stat finfo;
		stat (argv[1], &finfo);
		if (finfo.st_mode & S_IFDIR) {
			sw_dircomp = 1;
			Directoryname = argv [1];
			ArchiveFilename = new char [strlen (Directoryname) + 4];
			strcpy (ArchiveFilename, Directoryname);
			strcat (ArchiveFilename, ".rff");
		}
		else {
			sw_unc = 1;
			ArchiveFilename = argv [1];
		}
	}

	if (sw_unc || sw_list) deflate ();
	if (sw_comp) inflate ();
	if (sw_add) add ();
	if (sw_dircomp) inflatedir ();
	return 0;
}
